import unittest
from tmh.beskow import phonemap

class TestStringMethods(unittest.TestCase):

    def test_phonemap(self):
        pm = phonemap.phonemap('STA','IPA')
        self.assertEqual('foo'.upper(), 'FOO')


if __name__ == '__main__':
    unittest.main()



pm = phonemap.phonemap('STA','IPA')



"""
Utils handy for speech processing etc
"""
from io import BytesIO
import librosa
import torch
import soundfile as sf
import torchaudio
from scipy.io import wavfile
import noisereduce as nr


class ConversionError(Exception):
    """
    Exception for when conversion fails
    """
    pass


def change_sample_rate(audio_path: str, new_sample_rate: int = 16000):
    """
    Change the sample rate of an audio file. Defaults to 16000 Hz.
    """
    audio_to_resample, sr = librosa.load(audio_path)
    resampled_audio = librosa.resample(audio_to_resample, sr, new_sample_rate)
    resampled_tensor = torch.tensor([resampled_audio])
    return resampled_tensor

def ensure_sample_rate(audio_path: str, sample_rate: int, new_sample_rate: int = 16000):
    """
    Ensure that an audio file is in the specified sample rate. If not, convert it.
    Returns the path to the wav file as well as a boolean specifying if
    the file was converted.
    """
    if sample_rate != new_sample_rate:
        # resample to 16000 Hz
        waveform = change_sample_rate(
            audio_path, new_sample_rate=new_sample_rate)
    else:
        waveform, _ = torchaudio.load(audio_path)
    return waveform

def load_audio(audio_path: str, sample_rate: int = 16000):
    waveform, orig_sample_rate = torchaudio.load(audio_path)
    waveform = ensure_sample_rate(
        audio_path, orig_sample_rate, new_sample_rate=sample_rate)
    return waveform

def caf_bytes_to_wav(caf_bytes, file="test.wav"):
    audio, sr = librosa.load(BytesIO(caf_bytes))
    sf.write(file, audio, sr, subtype='PCM_24')

def convert_to_wav(audio_path: str, output_path: str = None):
    """
    Convert an audio file to wav format. Should be able to handle most file
    types, i.e. mp3, flac, caf, m4a, etc.

    audio_path: path to the audio file
    output_path: path to the output file (ending with .wav)
    """
    path = audio_path.split("/")
    filename = path[-1].split(".")[0]
    audio, sr = librosa.load(audio_path)
    wavpath = f'{filename}.wav'
    if output_path:
        wavpath = f"{output_path}/{wavpath}"
    sf.write(wavpath, audio, sr, subtype='PCM_24')
    print(f"Converted {audio_path} to wav:", wavpath)
    return wavpath


def ensure_wav(audio_path: str, reduce_noise: bool = False):
    """
    Ensure that an audio file is in wav format. If not, convert it.
    Returns the path to the wav file as well as a boolean specifying if
    the file was converted.
    """
    converted = False
    if not audio_path.endswith(".wav"):
        try:
            audio_path = convert_to_wav(audio_path)
            converted = True
        except FileNotFoundError as e:
            raise FileNotFoundError(e)
        except:
            raise ConversionError(f"Could not convert {audio_path} to wav")

    if reduce_noise:
        audio_path = reduce_noise(audio_path)

    return audio_path, converted


def reduce_noise(self, audio_path):
    print(f"Reducing noise for {audio_path}")
    # load data
    rate, data = wavfile.read(audio_path)
    # perform noise reduction
    reduced_noise = nr.reduce_noise(y=data, sr=rate)
    wavfile.write(audio_path, rate, reduced_noise)
    return audio_path



from time import sleep
from random import random
from multiprocessing import Process
from multiprocessing import Queue

# generate work


def producer(queue):
    print('Producer: Running', flush=True)
    # generate work
    for i in range(10):
        # generate a value
        value = random()
        # block
        sleep(value)
        # add to the queue
        queue.put(value)
    # all done
    queue.put(None)
    print('Producer: Done', flush=True)

# consume work


def consumer(queue):
    print('Consumer: Running', flush=True)
    # consume work
    while True:
        # get a unit of work
        item = queue.get()
        # check for stop
        if item is None:
            break
        # report
        print(f'>got {item}', flush=True)
    # all done
    print('Consumer: Done', flush=True)


# entry point
if __name__ == '__main__':
    # create the shared queue
    queue = Queue()
    # start the consumer
    consumer_process = Process(target=consumer, args=(queue,))
    consumer_process.start()
    # start the producer
    producer_process = Process(target=producer, args=(queue,))
    producer_process.start()
    # wait for all processes to finish
    producer_process.join()



import io
import torch
from transformers import Wav2Vec2ForCTC, Wav2Vec2ProcessorWithLM
import os
from tmh.transcribe_with_vad import extract_speak_segments
from tmh.utils import load_audio, ensure_wav
import soundfile as sf
import librosa

# from language_files import get_model


def transcribe_from_audio_path_with_lm(audio_path, model_id="viktor-enzell/wav2vec2-large-voxrex-swedish-4gram", model=None, processor=None):
    audio_path, converted = ensure_wav(audio_path)

    sample_rate = 16000
    waveform = load_audio(audio_path, sample_rate)

    if converted:
        os.remove(audio_path)

    if not (model and processor):
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        model = Wav2Vec2ForCTC.from_pretrained(model_id).to(device)
        processor = Wav2Vec2ProcessorWithLM.from_pretrained(model_id)
    else:
        device = model.device

    inputs = processor(waveform[0], sampling_rate=16000,
                       return_tensors='pt', padding=True).to(device)

    with torch.no_grad():
        logits = model(**inputs).logits

    transcripts = processor.batch_decode(logits.cpu().numpy()).text
    # print(transcripts)
    return transcripts[0]


def transcribe_from_audio_path_with_lm_vad(audio_path, model_id="viktor-enzell/wav2vec2-large-voxrex-swedish-4gram", model=None, processor=None, output_format='str'):
    audio_path, converted = ensure_wav(audio_path)

    sample_rate = 16000
    waveform = load_audio(audio_path, sample_rate)

    segments = extract_speak_segments(audio_path)
    transcriptions = []

    if not (model and processor):
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        model = Wav2Vec2ForCTC.from_pretrained(model_id).to(device)
        processor = Wav2Vec2ProcessorWithLM.from_pretrained(model_id)
    else:
        device = model.device

    for segment in segments['content']:
        x = waveform[:, int(segment['segment']['start']*sample_rate): int(segment['segment']['end']*sample_rate)]

        inputs = processor(x[0], sampling_rate=sample_rate,
                           return_tensors='pt', padding=True).to(device)

        with torch.no_grad():
            logits = model(**inputs).logits

        transcription = processor.batch_decode(logits.cpu().numpy()).text
        full_transcript = {
            "transcription": transcription[0].encode('utf8').decode().lower(),
            "start": segment['segment']['start'],
            "end": segment['segment']['end']
        }
        # print(transcription)
        transcriptions.append(full_transcript)

    if converted:
        os.remove(audio_path)

    if output_format == "str_dots":
        result = ". ".join([t["transcription"] for t in transcriptions])

    elif output_format == "str":
        result = " ".join([t["transcription"] for t in transcriptions])

    elif output_format == "json":
        result = transcriptions
    else:
        raise ValueError("Unknown output format")

    return result


def transcribe_bytes_with_lm(bytes, model_id="viktor-enzell/wav2vec2-large-voxrex-swedish-4gram", model=None, processor=None):
    # waveform, samplerate = sf.read(io.BytesIO(bytes))
    # audio, sr = librosa.load(io.BytesIO(bytes))
    # waveform, samplerate = sf.write(bytes, subtype='PCM_24')
    samplerate = bytes.samplerate
    waveform = bytes.read()

    if not (model and processor):
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        model = Wav2Vec2ForCTC.from_pretrained(model_id).to(device)
        processor = Wav2Vec2ProcessorWithLM.from_pretrained(model_id)
    else:
        device = model.device

    inputs = processor(waveform[0], sampling_rate=samplerate,
                       return_tensors='pt', padding=True).to(device)

    with torch.no_grad():
        logits = model(**inputs).logits

    transcripts = processor.batch_decode(logits.cpu().numpy()).text
    # print(transcripts)
    return transcripts[0]


def transcribe_bytes_with_lm_vad(bytes, model=None, processor=None, model_id="viktor-enzell/wav2vec2-large-voxrex-swedish-4gram", output_format='json'):
    waveform, sample_rate = sf.read(io.BytesIO(bytes))

    # TODO ensure that this is not a problem
    segments = extract_speak_segments(bytes)
    transcriptions = []

    if not (model and processor):
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        model = Wav2Vec2ForCTC.from_pretrained(model_id).to(device)
        processor = Wav2Vec2ProcessorWithLM.from_pretrained(model_id)
    else:
        device = model.device

    for segment in segments['content']:
        x = waveform[:, int(segment['segment']['start']*sample_rate): int(segment['segment']['end']*sample_rate)]

        inputs = processor(x[0], sampling_rate=sample_rate,
                           return_tensors='pt', padding=True).to(device)

        with torch.no_grad():
            logits = model(**inputs).logits

        transcription = processor.batch_decode(logits.cpu().numpy()).text
        full_transcript = {
            "transcription": transcription[0].encode('utf8').decode().lower(),
            "start": segment['segment']['start'],
            "end": segment['segment']['end']
        }
        # print(transcription)
        transcriptions.append(full_transcript)

    if output_format == "str_dots":
        result = ". ".join([t["transcription"] for t in transcriptions])

    elif output_format == "str":
        result = " ".join([t["transcription"] for t in transcriptions])

    elif output_format == "json":
        result = transcriptions
    else:
        raise ValueError("Unknown output format")

    return result



import phonemap
import sys

# mapdict - do phoneset maping on a tab-separated phonetic dictionary
# each row is assumed to be 'word<TAB>transcription'

# synopsis:
# python mapdict.py infile ?outfile?

from_bet = 'STA'
to_bet = 'MTM'

pm = phonemap.phonemap(from_bet,to_bet,mapfile='phonemap.tsv')


try:
    f = open(sys.argv[2],'w')
except:
    f = sys.stdout

for line in open(sys.argv[1]):
    txt,trans = line.strip().split('\t')
    newtrans = pm.map(pm.splitphonestring(trans))
    f.write('{}\t{}\n'.format(txt,' '.join(newtrans)))


import asyncio
import io
import logging
import multiprocessing
import torchaudio
import torch
from transformers import Wav2Vec2Processor, Wav2Vec2ForCTC, Wav2Vec2ProcessorWithLM
from transformers import HubertForSequenceClassification, Wav2Vec2FeatureExtractor
import librosa
from tmh.utils import ensure_wav
from tmh.language_files import get_model
from tmh.transcribe_with_vad import transcribe_bytes_split_on_speech, transcribe_from_audio_path_split_on_speech
from tmh.transcribe_with_lm import transcribe_bytes_with_lm, transcribe_bytes_with_lm_vad, transcribe_from_audio_path_with_lm, transcribe_from_audio_path_with_lm_vad

from speechbrain.pretrained import EncoderClassifier
import soundfile as sf
import os
import numpy as np

logger = logging.getLogger()


class ConversionError(Exception):
    pass


class TranscriptionError(Exception):
    pass


# TODO
# check language
# enable batch mode

PROCESSES = multiprocessing.cpu_count() - 1


class TranscribeModel:
    def __init__(self, use_vad=False, use_lm=False, language='Swedish', model_id=None):
        """
        use_vad: use voice activity detection
        use_lm: use language model
        language: language to use
        model_id: the name of a HuggingFace model, overrides `language`

        return: a TranscribeModel object
        """
        self.language = language
        self.use_vad = use_vad
        self.use_lm = use_lm
        self.model_id = model_id if model_id else self.get_model_id()
        # self.task_queue = multiprocessing.Queue() # TODO
        self.processes = []
        self.device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu')
        self.model, self.processor = self.initialize()
        # self.run() # TODO

    def run(self):
        while True:
            while len(self.processes) < PROCESSES:
                self.processes.append(
                    multiprocessing.Process(target=self.worker, kwargs={}))
                self.processes[-1].start()

    def worker(self, **kwargs):
        self.transcribe(**kwargs)

    def initialize(self):
        if self.use_vad and self.use_lm:
            self.processor = Wav2Vec2ProcessorWithLM.from_pretrained(
                self.model_id)
            self.model = Wav2Vec2ForCTC.from_pretrained(
                self.model_id).to(self.device)
            logger.info("Using LM+VAD")
        elif self.use_vad:
            self.processor = Wav2Vec2Processor.from_pretrained(self.model_id)
            self.model = Wav2Vec2ForCTC.from_pretrained(
                self.model_id).to(self.device)
            logger.info("Using VAD")
        elif self.use_lm:
            self.processor = Wav2Vec2ProcessorWithLM.from_pretrained(
                self.model_id)
            self.model = Wav2Vec2ForCTC.from_pretrained(
                self.model_id).to(self.device)
            logger.info("Using LM")
        else:
            self.processor = Wav2Vec2Processor.from_pretrained(self.model_id)
            self.model = Wav2Vec2ForCTC.from_pretrained(
                self.model_id).to(self.device)
        logger.info(
            "Model was initialized with model_id: {}".format(self.model_id))
        return self.model, self.processor

    def get_model_id(self):
        logger.info("getting model id")
        if self.use_lm:
            if self.language == "Swedish":
                logger.info("Using Swedish LM")
                return "viktor-enzell/wav2vec2-large-voxrex-swedish-4gram"
            else:
                logger.error(
                    "Language not supported for LM: {}".format(self.language))
                raise ValueError(
                    "Language not supported for LM: {}".format(self.language))
        else:
            return get_model(self.language)

    def transcribe(
            self,
            audio_path,
            classify_emotion=False,
            output_word_offsets=False,
            save_to_file=False,
            reduce_noise=False,
            output_format: str = "json"):
        """
        audio_path: path to audio file
        classify_emotion: whether to classify emotion or not (not yet implemented) # TODO
        output_word_offsets: whether to output word offsets or not (not yet implemented) # TODO
        save_to_file: whether to save to file or not
        reduce_noise: whether to reduce noise or not
        output_format: the format of the output, either "json", "str" or "str_dots"

        return: transcriptions in the format specified by output_format (default: json)
        """
        logger.info("Transcribing {}".format(audio_path))
        try:
            if self.use_vad and self.use_lm:
                logger.info("Using LM+VAD")
                return transcribe_from_audio_path_with_lm_vad(
                    audio_path=audio_path,
                    model=self.model,
                    processor=self.processor,
                    output_format=output_format)
            elif self.use_vad:
                logger.info("Using VAD")
                return transcribe_from_audio_path_split_on_speech(
                    audio_path,
                    save_to_file=save_to_file,
                    output_format=output_format,
                    model=self.model,
                    processor=self.processor)
            elif self.use_lm:
                logger.info("Using LM")
                return transcribe_from_audio_path_with_lm(
                    audio_path, model=self.model, processor=self.processor).lower()
            else:
                logger.info("Using neither LM nor VAD")
                return transcribe_from_audio_path(
                    audio_path=audio_path,
                    model=self.model,
                    processor=self.processor,
                    reduce_noise=reduce_noise,
                    classify_emotion=classify_emotion,
                    output_word_offsets=output_word_offsets)
        except Exception as e:
            logger.error(e)
            raise TranscriptionError(e)

    def queue_transcription(self, audio_path, *args):
        self.task_queue.put((audio_path, *args))

    def transcribe_bytes(self, bytes, output_format: str = "json"):
        """
        Transcribes a chunk of speech.

        @param speech: a chunk of speech with sample rate 16000 Hz

        @return: transcriptions in the format specified by output_format (default: json | text)
        """
        if self.use_vad and self.use_lm:
            return transcribe_bytes_with_lm_vad(
                bytes=bytes,
                model=self.model,
                processor=self.processor,
                output_format=output_format)
        elif self.use_vad:
            return transcribe_bytes_split_on_speech(
                bytes=bytes,
                model=self.model,
                processor=self.processor,
                output_format=output_format)
        elif self.use_lm:
            return transcribe_bytes_with_lm(
                bytes=bytes,
                model=self.model,
                processor=self.processor).lower()
        else:
            return transcribe_bytes(
                bytes=bytes,
                model=self.model,
                processor=self.processor)

    def process_tasks(self):
        logger = multiprocessing.get_logger()
        proc = os.getpid()
        while not self.task_queue.empty():
            try:
                (audio, *args) = self.task_queue.get()
                self.transcribe(audio, *args)
            except Exception as e:
                logger.error(e)
            logger.info(f"Process {proc} completed successfully")
        return True


def extract_speaker_embedding(audio_path):
    classifier = EncoderClassifier.from_hparams(
        source="speechbrain/spkrec-xvect-voxceleb", savedir="pretrained_models/spkrec-xvect-voxceleb")
    signal, fs = torchaudio.load(audio_path)
    embeddings = classifier.encode_batch(signal)
    # print(embeddings)
    return embeddings


def classify_emotion(audio_path):
    audio_path, converted = ensure_wav(audio_path)

    model = HubertForSequenceClassification.from_pretrained(
        "superb/hubert-large-superb-er")
    feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained(
        "superb/hubert-large-superb-er")
    speech, _ = librosa.load(audio_path, sr=16000, mono=True)

    inputs = feature_extractor(
        speech, sampling_rate=16000, padding=True, return_tensors="pt")

    logits = model(**inputs).logits
    predicted_ids = torch.argmax(logits, dim=-1)
    labels = [model.config.id2label[_id] for _id in predicted_ids.tolist()]
    # print(labels)
    if converted:
        os.remove(audio_path)
    return(labels)


def classify_language(audio_path):
    classifier = EncoderClassifier.from_hparams(
        source="speechbrain/lang-id-commonlanguage_ecapa", savedir="pretrained_models/lang-id-commonlanguage_ecapa")
    out_prob, score, index, text_lab = classifier.classify_file(audio_path)
    return(text_lab[0])


def get_speech_rate_time_stamps(time_stamps, downsample=320, sample_rate=16000):

    utterances = len(time_stamps[0])
    start_time = time_stamps[0][0]['start_offset']
    end_time = time_stamps[0][utterances-1]['end_offset']
    duration = end_time - start_time

    speech_rate = ((duration / utterances) * downsample) / sample_rate

    return speech_rate


def calculate_variance(data):
    n = len(data)
    mean = sum(data) / n
    # Square deviations
    deviations = [(x - mean) ** 2 for x in data]
    # Variance
    variance = sum(deviations) / n
    return variance


def get_speech_rate_variability(time_stamps, type='char', downsample=320, sample_rate=16000):
    base = downsample / sample_rate
    token_durations = {}

    for time_stamp in time_stamps[0]:

        start_time = round(time_stamp['start_offset']*base, 2)
        end_time = round(time_stamp['end_offset']*base, 2)
        char = time_stamp[type]
        duration = end_time - start_time

        if char not in token_durations:
            token_durations[char] = []

        token_durations[char].append(duration)

    averages = dict()
    stds = dict()
    variances = dict()

    for token, durations in token_durations.items():
        average = np.sum(durations) / len(durations)
        std = np.std(durations)
        # print("the tokens are", token)
        # print("the durations are", durations)
        # print("the average is", average)
        variance = calculate_variance(durations)
        averages[token] = average
        stds[token] = std
        variances[token] = variance

    return averages, stds, variances


def transcribe_from_audio_path(audio_path, model=None, processor=None, model_id=None, language='Swedish', check_language=False, reduce_noise=False, classify_emotion=False, output_word_offsets=False):
    audio_path, converted = ensure_wav(audio_path, reduce_noise=reduce_noise)

    sample_rate = 16000

    if not (model and processor) and not model_id:
        if check_language:
            language = classify_language(audio_path)
        # print("the language is", language)
        model_id = get_model(language)

    if not (model and processor):
        device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu')
        processor = Wav2Vec2Processor.from_pretrained(model_id)
        model = Wav2Vec2ForCTC.from_pretrained(model_id).to(device)
    else:
        device = model.device

    transcript = ""
    # Ensure that the sample rate is 16k

    # Stream over 30 seconds chunks rather than load the full file
    stream = librosa.stream(
        audio_path,
        block_length=30,
        frame_length=sample_rate,
        hop_length=sample_rate
    )

    for speech in stream:
        if len(speech.shape) > 1:
            speech = speech[:, 0] + speech[:, 1]

        input_values = processor(
            speech, sampling_rate=sample_rate, return_tensors="pt").input_values.to(device)
        logits = model(input_values).logits

        predicted_ids = torch.argmax(logits, dim=-1)
        transcription = processor.decode(predicted_ids[0])
        transcript += transcription.lower()
        # print(transcription[0])

    if converted:
        os.remove(audio_path)

    return transcript


def transcribe_bytes(bytes, model=None, processor=None, model_id=None, language='Swedish'):
    speech, sample_rate = sf.load(io.BytesIO(bytes))

    if not (model and processor) and not model_id:
        # print("the language is", language)
        model_id = get_model(language)

    if not (model and processor):
        device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu')
        processor = Wav2Vec2Processor.from_pretrained(model_id)
        model = Wav2Vec2ForCTC.from_pretrained(model_id).to(device)
    else:
        device = model.device

    transcript = ""
    # Ensure that the sample rate is 16k

    if len(speech.shape) > 1:
        speech = speech[:, 0] + speech[:, 1]

    input_values = processor(
        speech, sampling_rate=sample_rate, return_tensors="pt").input_values.to(device)
    logits = model(input_values).logits

    predicted_ids = torch.argmax(logits, dim=-1)
    transcription = processor.decode(predicted_ids[0])
    transcript += transcription.lower()
    # print(transcription[0])

    return transcript


def output_word_offset(pred_ids, processor, output_word_offsets):
    outputs = processor.batch_decode(
        pred_ids, output_word_offsets=output_word_offsets)
    transcription = outputs["text"][0]
    token_time_stamps = outputs[1]
    speech_rate = get_speech_rate_time_stamps(token_time_stamps)
    averages, stds, variances = get_speech_rate_variability(
        token_time_stamps, type="char")
    word_time_stamps = outputs[2]
    return {
        "transcription": transcription,
        "speech_rate": speech_rate,
        "averages": averages,
        "standard_deviations": stds,
        "variances": variances
    }


if __name__ == "__main__":
    path = "word1.wav"
    t = TranscribeModel(use_lm=True)
    transcript = t.transcribe(path)
    print(transcript)



from tmh.transcribe import TranscribeModel
from tmh.utils import ensure_wav

file, converted = ensure_wav("tmh/tests/word1.m4a")

normal = TranscribeModel()
norm_t = normal.transcribe(file)
print("Normal:", norm_t)

vad = TranscribeModel(use_vad=True)
vad_t = vad.transcribe(file, output_format="str")
print("VAD:", vad_t)

lm = TranscribeModel(use_lm=True)
lm_t = lm.transcribe(file)
print("LM:", lm_t)

lm_vad = TranscribeModel(use_lm=True, use_vad=True)
lm_vad_t = lm_vad.transcribe(file, output_format="str")
print("LM+VAD:", lm_vad_t)



# from vad import extract_speak_segments
from librosa.core.spectrum import power_to_db
import torch
from .overlap import overlap_detection
from .transcribe import extract_speaker_embedding
import numpy as np

from sklearn.preprocessing import StandardScaler

from pydub import AudioSegment
from .audio_embeddings import get_audio_embeddings

from pyannote.audio import Inference
from pyannote.audio.pipelines import VoiceActivityDetection # pip install https://github.com/pyannote/pyannote-audio/archive/develop.zip
from transformers import Wav2Vec2Processor, Wav2Vec2Model
from sklearn.mixture import GaussianMixture
from sklearn.cluster import KMeans, AgglomerativeClustering, SpectralClustering, MiniBatchKMeans
from pyannote.core import Segment, Timeline
import torchaudio
from speechbrain.pretrained import EncoderClassifier
# from language_files import get_model

pipeline = VoiceActivityDetection(segmentation="pyannote/segmentation")

# speaker embedding
inference = Inference("pyannote/embedding", window="whole")

HYPER_PARAMETERS = {
  # onset/offset activation thresholds
  "onset": 0.5, "offset": 0.5,
  # remove speech regions shorter than that many seconds.
  "min_duration_on": 0.0,
  # fill non-speech regions shorter than that many seconds.
  "min_duration_off": 0.0
}

def extract_speak_segments(audio_path):
    pipeline.instantiate(HYPER_PARAMETERS)    
    vad = pipeline(audio_path)
    # print("extracting speaker segments")
    # print(vad)
    return(vad.for_json())


def classify_speakers_based_on_embeddings(X):
    X = StandardScaler().fit_transform(X)
    
    estimator=GaussianMixture(n_components=2)
    y = estimator.fit_predict(X)
    return y


def extract_speaker_embedding(signal, audio):

    audio_file = torch.tensor(audio[signal.start*1000 : signal.end * 1000].get_array_of_samples())
    #embeddings = classifier.encode_batch(audio_file)
    embeddings = inference( {"waveform": audio_file.unsqueeze(0).float(), "sample_rate": 44100} )

    # print(embeddings)
    return embeddings
    


def create_speaker_files_from_audio_path(audio_path):

    from pyannote.audio import Pipeline
    pipeline = Pipeline.from_pretrained("pyannote/speaker-diarization")

    # apply pretrained pipeline
    audio = AudioSegment.from_wav(audio_path)

    diarization = pipeline(audio_path)

    speaker1= AudioSegment.empty()
    speaker2 = AudioSegment.empty()

    # print the result
    for turn, _, speaker in diarization.itertracks(yield_label=True):
        print(f"start={turn.start:.1f}s stop={turn.end:.1f}s speaker_{speaker}")
        current_audio_segment = audio[turn.start*1000:turn.end*1000]
        if speaker == "SPEAKER_00":
            print("speaker1", speaker)
            speaker1+=current_audio_segment
        elif speaker == "SPEAKER_01":
            print("speaker2", speaker2)
            speaker2+=current_audio_segment

        speaker1.export("speaker1.wav", format="wav")
        speaker2.export("speaker2.wav", format="wav")


def wavlm_speaker_diarization(audio_path):
    from transformers import Wav2Vec2FeatureExtractor, UniSpeechSatForAudioFrameClassification
    import torch

    audio_file = torchaudio.load(audio_path)

    feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained('microsoft/wavlm-base-sd')
    model = UniSpeechSatForAudioFrameClassification.from_pretrained('microsoft/wavlm-base-sd')



    # audio file is decoded on the fly
    inputs = feature_extractor(audio_file, return_tensors="pt")
    logits = model(**inputs).logits
    probabilities = torch.sigmoid(logits[0])

    # labels is a one-hot array of shape (num_frames, num_speakers)
    labels = (probabilities > 0.5).long()
    print(labels)


def create_speaker_files_from_audio_path_old(audio_path):
    """
    Creates speaker files from the audio file at audio_path by splitting based on speaker.

    If the 'save_to_file' parameter is set to a file name, the
    results will also be written to file. 

    The srt format can be used to create accurately timed subtitles 
    for videos. If 'audio_path' is the audio track of a video file, 
    a video player like VLC will create those subtitles automatically, 
    given a file in the srt format.
    """
    
    audio = AudioSegment.from_wav(audio_path)

    if audio.frame_rate != 16000:
        ## change sample rate to 16000
        waveform = audio.set_frame_rate(16_000)

    
    # first we get all the overlaps for each speaker in the file
    overlaps = overlap_detection(audio_path)
    print(overlaps)

    # removing overlaps
    for overlapping_segment in overlaps:
        print(overlapping_segment  )


    starting_time = 0
    without_overlaps_audio = AudioSegment.empty()
    for overlaping_segments in overlaps['content']:
        segment = overlaping_segments['segment']
        current_segment_start_time = segment['start'] * 1000
        current_segment_end_time = segment['end'] * 1000
        current_segment = audio[starting_time: current_segment_start_time]
        starting_time = current_segment_end_time
        
        without_overlaps_audio += current_segment
    
    without_overlaps_audio.export("temp.wav", format="wav")
    
    segments = extract_speak_segments('temp.wav')
 
    embeddings = []
        
    inference = Inference("pyannote/embedding", window="whole")
    
    timeline = Timeline()
    count_short = 0
    for segment in segments['content']:
        current_segment_time = segment['segment']
        start_time = current_segment_time['start']
        end_time = current_segment_time['end']

        if end_time - start_time < 1:
            count_short += 1
            print(str(count_short) + ' short segments')
            continue
        current_segment = Segment(start_time, end_time)
        print('Long segment: ', current_segment)
        #embedding = inference.crop("temp.wav", current_segment)
        embedding = extract_speaker_embedding(current_segment, audio)
        
        
        timeline.add(current_segment)
        embeddings.append(embedding)
    
    embeddings = np.vstack(embeddings)
    
    y = classify_speakers_based_on_embeddings(embeddings)

    speaker1= AudioSegment.empty()
    speaker2 = AudioSegment.empty()

    for i, segment in enumerate(timeline):
        current_audio_segment = audio[segment.start*1000:segment.end*1000]
        if y[i] == 0:


            speaker1 += current_audio_segment
        else:
            speaker2 += current_audio_segment

    
    speaker1.export("speaker1_old.wav", format="wav")
    speaker2.export("speaker2_old.wav", format="wav")
    
            

    return embeddings

# model_ids = {
#         "hubert": "facebook/hubert-large-ls960-ft",
#         "wav2vec2": "facebook/wav2vec2-base-960h",
# }

# processor = Wav2Vec2Processor.from_pretrained("facebook/wav2vec2-base-960h")
# model = Wav2Vec2Model.from_pretrained("facebook/wav2vec2-base-960h")

# def get_audio_embeddings(audio_array, sample_rate, model_id="facebook/wav2vec2-base-960h"):
#     input_values = processor(audio_array, sampling_rate=sample_rate, return_tensors="pt").input_values
#     output = model(**input_values)
#     hidden_states = output.last_hidden_state
#     return hidden_states


def time_format( t ) :
    """
    Produces the time format expected by the srt format,
    given a parameter t representing a number of seconds.
    """
    hours = int(t/3600)
    minutes = int((t-hours*3600)/60)
    seconds = int(t-hours*3600-minutes*60)
    fraction = int((t%1)*100)
    return str(hours) + ":" + str(minutes) + ":" + str(seconds) + "." + str(fraction)





import csv
import os
class phonemap:

    def __init__(self,frombet,tobet,mapfile=None,stripchars='',accept_empty=False,verbose=False):
        self.m = {}
        self.charmap = {}
        self.accept_empty = accept_empty
        self.stripchars = stripchars
        self.verbose = verbose
        dir = os.path.dirname(os.path.abspath(__file__))
        if mapfile:
            self.mapfile = mapfile
        else:
            self.mapfile = os.path.join(dir,'phonemap.tsv')
        with open(self.mapfile) as csvf:
            dictreader = csv.DictReader(csvf,delimiter='\t')
            if not frombet in dictreader.fieldnames:
                raise Exception('frombet {} not in mapfile'.format(frombet))
            if not tobet in dictreader.fieldnames:
                raise Exception('tobet {} not in mapfile'.format(tobet))

            for item in dictreader:
                if tobet.startswith('is'):
                    self.m[item[frombet]] = bool(eval(item[tobet]))  #interpret numbers and 'True'/'False' as boolean
                else:
                    self.m[item[frombet]] = item[tobet]
    def normalize(self,sym):
        sym = sym.replace('ː',':').replace('ɡ','g')
        if self.stripchars:
            sym = sym.strip(self.stripchars)
        return sym

    def map(self,item):

        if type(item)==list: # list of symbols
            return [self.map(p) for p in item]
        else: # single symbol
            sym = self.normalize(item)
            ret = self.m[sym]
            if sym in self.m:
                if self.verbose:
                    print(sym,'->',ret)
                return ret
            else:
                print('no match for',sym)
                if self.accept_empty:
                    return ''
                else:
                    raise Exception('no mapping for {}'.format(sym))
    
    def splitphonestring(self,string):
        allphones = sorted(list(self.m),key=len,reverse=True) # all phones sorted by length
        allphones = [p for p in allphones if p!='']
        maxlen = len(allphones[0])
        res = []
        string = string.strip()
        smatch = False
        for ph in allphones:
            if string.startswith(ph):
                #print(ph)
                smatch = True
                res.append(ph)
                rest = string[len(ph):]
                #print(res,'-',rest)
                if rest in allphones:
                    res.append(rest)
                elif len(rest)>1:
                    res += self.splitphonestring(rest)
                else:
                    raise Exception('?:'+rest)
                break
        if not smatch:
            raise Exception('no match for:'+string)
        return res

    def exists(self,sym):
        sym = self.normalize(sym)
        return sym in self.m

if __name__ == '__main__':

    #    pm = phonemap('MTM','IPA','swmap.tsv')
    pm = phonemap('STA','IPA')
    s = 'A:'
    print('map():\n',s,'->',pm.map(s))
    a = 'SJÖ:SJU:KA'
    b = pm.splitphonestring(a)
    print('splitphonestring():\n',a,'\n',b)
    print('map():\n',b,'\n',pm.map(b))
